"""Not a dummer any more 
"""
import torch
import numpy as np 

from fractal.anchors.utils import bbox_iou, xywh_xyxy, clamper, xyxy_xywh
from fractal.anchors.anchors_generator import generate_anchors



def bbox_iou(box1, box2, x1y1x2y2=True):
    if x1y1x2y2:
        x1_min = min(box1[0], box2[0])
        x2_max = max(box1[2], box2[2])
        y1_min = min(box1[1], box2[1])
        y2_max = max(box1[3], box2[3])
        w1, h1 = box1[2] - box1[0], box1[3] - box1[1]
        w2, h2 = box2[2] - box2[0], box2[3] - box2[1]
    else:
        w1, h1 = box1[2], box1[3]
        w2, h2 = box2[2], box2[3]
        x1_min = min(box1[0]-w1/2.0, box2[0]-w2/2.0)
        x2_max = max(box1[0]+w1/2.0, box2[0]+w2/2.0)
        y1_min = min(box1[1]-h1/2.0, box2[1]-h2/2.0)
        y2_max = max(box1[1]+h1/2.0, box2[1]+h2/2.0)

    w_union = x2_max - x1_min
    h_union = y2_max - y1_min
    w_cross = w1 + w2 - w_union
    h_cross = h1 + h2 - h_union
    carea = 0
    if w_cross <= 0 or h_cross <= 0:
        return 0.0

    area1 = w1 * h1
    area2 = w2 * h2
    carea = w_cross * h_cross
    uarea = area1 + area2 - carea
    return float(carea/uarea)

def multi_bbox_ious(boxes1, boxes2, x1y1x2y2=True):
    """ To find iou by 1to1 mapping
    (4, 507) * (4*507) --- > 507
    """
    if x1y1x2y2:
        x1_min = torch.min(boxes1[0], boxes2[0])
        x2_max = torch.max(boxes1[2], boxes2[2])
        y1_min = torch.min(boxes1[1], boxes2[1])
        y2_max = torch.max(boxes1[3], boxes2[3])
        w1, h1 = boxes1[2] - boxes1[0], boxes1[3] - boxes1[1]
        w2, h2 = boxes2[2] - boxes2[0], boxes2[3] - boxes2[1]
    else:
        w1, h1 = boxes1[2], boxes1[3]
        w2, h2 = boxes2[2], boxes2[3]
        x1_min = torch.min(boxes1[0]-w1/2.0, boxes2[0]-w2/2.0)
        x2_max = torch.max(boxes1[0]+w1/2.0, boxes2[0]+w2/2.0)
        y1_min = torch.min(boxes1[1]-h1/2.0, boxes2[1]-h2/2.0)
        y2_max = torch.max(boxes1[1]+h1/2.0, boxes2[1]+h2/2.0)

    w_union = x2_max - x1_min
    h_union = y2_max - y1_min
    w_cross = w1 + w2 - w_union
    h_cross = h1 + h2 - h_union
    mask = (((w_cross <= 0) + (h_cross <= 0)) > 0)
    area1 = w1 * h1
    area2 = w2 * h2
    carea = w_cross * h_cross
    carea[mask] = 0
    uarea = area1 + area2 - carea
    return carea/uarea

def process_yolo_output(output, stride, anchor_scales, anchors, nW, nH, nC):
    """ The most crazy function. Outputs the prediction for each box on feature scale
    """
    nB = output.shape[0]
    grid_size_x = int(nW/stride)
    grid_size_y = int(nH/stride)
    nAnchors = anchor_scales.shape[0]
    #print(nB, grid_size_x, grid_size_y, nAnchors)
    output = output.view(nB, (nC+5)*nAnchors, grid_size_x * grid_size_y).transpose(1, 2).contiguous()
    output = output.view(nB, grid_size_x * grid_size_y * nAnchors, nC+5)
    #print("before", output[:, :, 4].cpu().unique())
    
    output[:, :, :2] = torch.sigmoid(output[:, :, :2])
    output[:, :, 4] = torch.sigmoid(output[:, :, 4])
    output[:, :, 5:] = torch.sigmoid(output[:, :, 5:])
    #print("after", output[:, :, 4].cpu().unique())
    
    scores = output.clone()
    
    output[:, :, :2] += anchors[:, :2]
    output[:, :, 2]  = output[:, :, 2].exp() * anchors[:, 2]
    output[:, :, 3]  = output[:, :, 3].exp() * anchors[:, 3]
    return output, scores

def encode_targets(outputs, targets, anchor_scales, anchors, nW, nH, nC, strides, nthresh=0.7):
    eps = 0.00001
    nB = outputs.size(0)
    nA = outputs.size(1)
    nG = targets.shape[1]
    conf_mask = torch.ones(nB, nA)
    coord_mask = torch.zeros(nB, nA, 4)
    cls_mask = torch.zeros(nB, nA, nC)
    tcoord = torch.zeros(nB, nA , 4)
    tconf = torch.zeros(nB, nA)
    tcls = torch.zeros(nB, nA)
    twidth, theight = nW/strides, nH/strides
    anchors_xy = anchors.clone()
    anchors_xy[:, :2] += 0.5
    
    nGT=0
    nRecall = 0
    nRecall75 = 0
    
    for b in range(nB):
        label = targets[b]
        cur_pred_boxes = outputs[b]
        cur_ious = torch.zeros(nA)
        for la in label:
            if la[0] == 0:
                break
            gx, gy, gw, gh = la[1:]/strides
            cur_gt_boxes = torch.FloatTensor([gx, gy, gw, gh]).repeat(nA, 1)
            ious = multi_bbox_ious(cur_gt_boxes.t(), anchors_xy.t(), x1y1x2y2=False)
            cur_ious = torch.max(cur_ious, ious)
        ignore_ix = cur_ious > nthresh
        conf_mask[b, ignore_ix] = 0
        
        for la in label:
            if la[0] == 0:
                break
            nGT+=1
            gx, gy, gw, gh = la[1:]/strides
            tmp_gt_boxes = torch.FloatTensor([0, 0, gw, gh]).repeat(anchor_scales.shape[0], 1)
            bestn_start = ((anchors[:, 0] == int(gx)) & (anchors[:, 1] == int(gy))).nonzero()[0]
            k = anchors_xy[(anchors[:, 0] == int(gx)) & (anchors[:, 1] == int(gy))]
            k[:, :2] = 0
            values, bestn = torch.max(multi_bbox_ious(tmp_gt_boxes.t(), \
                                     k.t(), x1y1x2y2=False), 0)
            total_n = int(bestn+bestn_start)
            #assigned_anchor = anchors[total_n]
            gt_box = torch.FloatTensor([gx, gy, gw, gh])
            iou = bbox_iou(gt_box, outputs[b, total_n][:4].cpu(), x1y1x2y2=False)
            
            if iou > 0.5:
                nRecall +=1
                if iou > 0.75:
                    nRecall75 +=1
            
            coord_mask[b, total_n] = 1 
            cls_mask[b, total_n, int(la[0]) - 1] = 1
            conf_mask[b, total_n] = 5
            tcoord[b, total_n, 0] = float(gx - int(gx))
            tcoord[b, total_n, 1] = float(gy - int(gy))
            tcoord[b, total_n, 2] = float(torch.log((gw+eps)/anchors[total_n, 2]))
            tcoord[b, total_n, 3] = float(torch.log((gh+eps)/anchors[total_n, 3]))
            tcls[b, total_n] =  1
            tconf[b, total_n] = 1#iou if 0 else 1
    return coord_mask, conf_mask, cls_mask, tcoord, tconf, tcls, nRecall, nRecall75, nGT
            
            
    
    
    


import numpy as np 
import torch

def xyxy_xywh(boxes, offset=0.5):
    """xyxy to xywh format  
    """
    height = boxes[:, 3] - boxes[:, 1]
    width = boxes[:, 2] - boxes[:, 0]
    ctr_y = boxes[:, 1] + offset * height
    ctr_x = boxes[:, 0] + offset * width 
    return torch.stack((ctr_x, ctr_y, width, height), dim=1)

def clamper(bbox, size):
    """bbox takes xyxy values
    """
    h, w = size
    bbox[:, 0] = torch.clamp(bbox[:, 0], 0, w)
    bbox[:, 1] = torch.clamp(bbox[:, 1], 0, h)
    bbox[:, 2] = torch.clamp(bbox[:, 2], 0, w)
    bbox[:, 3] = torch.clamp(bbox[:, 3], 0, h)
    return bbox
    
def xywh_xyxy(boxes):
    """xyhw_xyxy ----> we are using h,w and not w, h format .. Please be careful
    """
    bbox = torch.zeros(boxes.shape)
    bbox[:, 0] = boxes[:, 0] - 0.5 * boxes[:, 2]
    bbox[:, 1] = boxes[:, 1] - 0.5 * boxes[:, 3]
    bbox[:, 2] = boxes[:, 0] + 0.5 * boxes[:, 2]
    bbox[:, 3] = boxes[:, 1] + 0.5 * boxes[:, 3]
    return bbox

def bbox_iou(bbox_a, bbox_b, intersection=True):
    """Calculate the Intersection of Unions (IoUs) between
    bounding boxes.
    IoU is calculated as a ratio of area of the intersection
    and area of the union.

    bbox_a: numpy array
    bbox_b: numpy array

    Returns the area between two numpy array's. (yxyx format)
    
    
    """
    #ToDo :Convert to Pytorch
    if bbox_a.shape[1] != 4 or bbox_b.shape[1] != 4:
        raise IndexError

    # top left
    tl = np.maximum(bbox_a[:, None, :2], bbox_b[:, :2])
    # bottom right
    br = np.minimum(bbox_a[:, None, 2:], bbox_b[:, 2:])

    area_i = np.prod(br - tl, axis=2) * (tl < br).all(axis=2)
    area_a = np.prod(bbox_a[:, 2:] - bbox_a[:, :2], axis=1)
    area_b = np.prod(bbox_b[:, 2:] - bbox_b[:, :2], axis=1)
    if intersection:
        iou = area_i / (area_a[:, None] + area_b - area_i)
    else:
        iou = area_i 
    return iou



import torch
from torchvision.transforms import transforms
import numpy as np
import visdom

import matplotlib.pyplot as plt 
from PIL import Image


class VisdomLinePlotter(object):
    """Plots to Visdom
    """
    def __init__(self, env_name='main'):
        self.viz = visdom.Visdom()
        self.env = env_name
        self.plots = {}
    def plot(self, var_name, split_name, x, y):
        """ Plot all the metrics available as named tuple
        
        var_name: named tuple 
        """
        if var_name not in self.plots:
            self.plots[var_name] = self.viz.line(X=np.array([x,x]), Y=np.array([y,y]), env=self.env, opts=dict(
                legend=[split_name],
                title=var_name,
                xlabel='Iterations',
                ylabel=var_name
            ))
        else:
            self.viz.line(X=np.array([x]), Y=np.array([y]), env=self.env, win=self.plots[var_name], name=split_name, update = 'append')

    def plot_tuple(self, ntuple, name, iteration):

        for field, value in ntuple._asdict().items():
            self.plot(field, name, iteration, value.data.cpu().numpy()[0])


import math
import random

import torch
import torchvision.transforms as transforms

from PIL import Image, ImageDraw
import random

__all__ = ["resize", "random_crop", "center_crop", "random_flip", "random_distort_image"]

def resize(img, size, boxes=None, max_size=1333):
    '''Resize the input PIL image to the given size.
    Args:
      img: (PIL.Image) image to be resized.
      boxes: (tensor) object boxes, sized [#ojb,4].
      size: (tuple or int)
        - if is tuple, resize image to the size.
        - if is int, resize the shorter side to the size while maintaining the aspect ratio.
      max_size: (int) when size is int, limit the image longer size to max_size.
                This is essential to limit the usage of GPU memory.
    Returns:
      img: (PIL.Image) resized image.
      boxes: (tensor) resized boxes.
    '''
    w, h = img.size
    if isinstance(size, int):
        size_min = min(w,h)
        size_max = max(w,h)
        sw = sh = float(size) / size_min
        if sw * size_max > max_size:
            sw = sh = float(max_size) / size_max
        ow = int(w * sw + 0.5)
        oh = int(h * sh + 0.5)
    else:
        ow, oh = size
        sw = float(ow) / w
        sh = float(oh) / h
    
    if boxes is None:
        return img.resize((ow, oh), Image.BILINEAR), 0
    else:
        return img.resize((ow,oh), Image.BILINEAR), boxes*torch.Tensor([sh,sw,sh,sw])



def random_crop(img, boxes):
    '''Crop the given PIL image to a random size and aspect ratio.
    A crop of random size of (0.08 to 1.0) of the original size and a random
    aspect ratio of 3/4 to 4/3 of the original aspect ratio is made.
    Args:
      img: (PIL.Image) image to be cropped.
      boxes: (tensor) object boxes, sized [#ojb,4].
    Returns:
      img: (PIL.Image) randomly cropped image.
      boxes: (tensor) randomly cropped boxes.
    '''
    success = False
    for attempt in range(10):
        area = img.size[0] * img.size[1]
        target_area = random.uniform(0.56, 1.0) * area
        aspect_ratio = random.uniform(3. / 4, 4. / 3)

        w = int(round(math.sqrt(target_area * aspect_ratio)))
        h = int(round(math.sqrt(target_area / aspect_ratio)))

        if random.random() < 0.5:
            w, h = h, w

        if w <= img.size[0] and h <= img.size[1]:
            x = random.randint(0, img.size[0] - w)
            y = random.randint(0, img.size[1] - h)
            success = True
            break

    # Fallback
    if not success:
        w = h = min(img.size[0], img.size[1])
        x = (img.size[0] - w) // 2
        y = (img.size[1] - h) // 2

    img = img.crop((x, y, x+w, y+h))
    boxes -= torch.Tensor([y,x,y,x])
    boxes[:,0::2].clamp_(min=0, max=h-1)
    boxes[:,1::2].clamp_(min=0, max=w-1)
    return img, boxes


def center_crop(img, boxes, size):
    '''Crops the given PIL Image at the center.
    Args:
      img: (PIL.Image) image to be cropped.
      boxes: (tensor) object boxes, sized [#ojb,4].
      size (tuple): desired output size of (w,h).
    Returns:
      img: (PIL.Image) center cropped image.
      boxes: (tensor) center cropped boxes.
    '''
    w, h = img.size
    ow, oh = size
    i = int(round((h - oh) / 2.))
    j = int(round((w - ow) / 2.))
    img = img.crop((j, i, j+ow, i+oh))
    boxes -= torch.Tensor([i,j,i,j])
    boxes[:,0::2].clamp_(min=0, max=oh-1)
    boxes[:,1::2].clamp_(min=0, max=ow-1)
    return img, boxes

def random_flip(img, boxes):
    '''Randomly flip the given PIL Image.
    Args:
        img: (PIL Image) image to be flipped.
        boxes: (tensor) object boxes, sized [#ojb,4].
    Returns:
        img: (PIL.Image) randomly flipped image.
        boxes: (tensor) randomly flipped boxes.
    '''
    if random.random() < 0.5:
        img = img.transpose(Image.FLIP_LEFT_RIGHT)
        w = img.width
        xmin = w - boxes[:,3]
        xmax = w - boxes[:,1]
        boxes[:,1] = xmin
        boxes[:,3] = xmax
    return img, boxes

def distort_image(im, hue, sat, val):
    im = im.convert('HSV')
    cs = list(im.split())
    cs[1] = cs[1].point(lambda i: i * sat)
    cs[2] = cs[2].point(lambda i: i * val)
    
    def change_hue(x):
        x += hue*255
        if x > 255:
            x -= 255
        if x < 0:
            x += 255
        return x
    cs[0] = cs[0].point(change_hue)
    im = Image.merge(im.mode, tuple(cs))

    im = im.convert('RGB')
    #constrain_image(im)
    return im

def rand_scale(s):
    scale = random.uniform(1, s)
    if(random.randint(1,10000)%2): 
        return scale
    return 1./scale

def random_distort_image(im, hue, saturation, exposure):
    dhue = random.uniform(-hue, hue)
    dsat = rand_scale(saturation)
    dexp = rand_scale(exposure)
    res = distort_image(im, dhue, dsat, dexp)
    return res


""" Schedulars to fix learning_rate, weight_decay, momentum and other stuff after each iteration. 
"""
import torch 
import torch.nn as nn 
import torch.nn.functional as F 
import math 
import numpy as np

from functools import partial

def get_schedulers(sched_name, optimizer, train_images, batch_size, kwargs=None):
    if sched_name is None:
        print(sched_name, kwargs)
        if kwargs is None:
            kwargs = {"lr": 0.001}
        return No_scheduler(optimizer, 
                            **kwargs)

    elif sched_name == "sgdr":
        if kwargs is None:
            kwargs = {"t_actual": 10, "t_mul": 2, "lr_max": 0.05, "lr_min": 0.0, "decay": 1}
        
        return Warm_restarts(optimizer,
                     batch_size, 
                     train_images,
                     **kwargs)

    elif sched_name == "one_cycle":
        if kwargs is None:
            kwargs = {"stepsize": 4, "max_lr": 1, "base_lr": 0.1, "max_mom": 0.9, "base_mom": 0.9, "epochs": 10}
        
        return one_cycle(optimizer, 
                     batch_size, 
                     train_images,
                     **kwargs)

    elif sched_name == "cyclic_lr":
        if kwargs is None:
            kwargs = {"stepsize": 5, "max_lr": 0.1, "base_lr": 0.02}

        return cyclic_lr(optimizer,
                     batch_size,
                     train_images,
                     **kwargs)

    elif sched_name == "ReduceLRonPlateau":
        if kwargs is None:
            kwargs = {"mode": 'min', "factor": 0.2, "patience": 5, "verbose": False, "threshold": 1e-4, "threshold_mode": 'rel', 
                       "cooldown" : 0, "min_lr": 0.0, "eps": 1e-8}
        return ReduceLRonPlateau(optimizer,
                     batch_size, 
                     train_images,
                     **kwargs)

    elif sched_name == "StepLR":
        if kwargs is None:
            kwargs = {"step_size": 10, "gamma": 0.5, "last_epoch": -1}

        return StepLR(optimizer,
                     batch_size,
                     train_images,
                     **kwargs)

    else:
        raise NotImplementedError("Schedular is not implemented, if there is need for it ... Pull request welcome")


class No_scheduler:
    def __init__(self, optimizer, lr):
        self.optimizer = optimizer
        self.lr = lr
        self.cycle = 0 
        self.last_batch_iteration = 0 
    
    def get_lr(self):
        return self.lr
    
    def fix_lr(self, *argv):
        new_lr = self.get_lr()

        for param_group in self.optimizer.param_groups:
            param_group["lr"] = new_lr
        return new_lr 
    
    def load_state_dict(self, save):
        self.optimizer.load_state_dict(save["optimizer"])
        self.last_batch_iteration = save["last_batch_iteration"]
        self.cycle = save["cycle"]
    
    def state_dict(self):
        save = {}
        save["optimizer"] = self.optimizer.state_dict()
        save["last_best_iteration"] = self.last_batch_iteration 
        save["cycle"] = self.cycle 
        return save 

        
class Warm_restarts:
    def __init__(self, optimizer, batch_size, total_images, t_actual, t_mul, lr_max, lr_min, decay):
        """ SGDR warm restarts scheduler 
        optimizer: optimizer used for training 
        t_actual: 
        """
        self.optimizer = optimizer
        self.t_actual = t_actual
        self.t_mul = t_mul
        self.lr_max = lr_max
        self.lr_min = lr_min
        self.decay = decay
        self.batch_size = batch_size 
        self.total_images = total_images
        self.last_batch_iteration = 0
        self.iterations_per_batch = math.ceil(self.total_images/self.batch_size)
        self.cycle = 0
        self.ti = self.iterations_per_batch * self.t_actual
    
    def get_lr(self):
        iteration = self.last_batch_iteration/self.ti
        cosine_value = math.cos(iteration * math.pi)
        new_lr = self.lr_min + (0.5 * (self.lr_max - self.lr_min) * (1+cosine_value))
        self.last_batch_iteration += 1
        
        if self.last_batch_iteration==self.ti:
            print("Warm restart")
            self.last_batch_iteration = 0
            self.ti *= self.t_mul
            self.cycle += 1 
            self.lr_max = self.lr_max/self.decay
        return new_lr
    
    def fix_lr(self, *argv):
        new_lr = self.get_lr()
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = new_lr
        return new_lr
    
    def load_state_dict(self, save):
        self.optimizer.load_state_dict(save["optimizer"])
        self.last_batch_iteration = save["last_batch_iteration"]
        self.cycle = save["cycle"]
    
    def state_dict(self):
        save = {}
        save["optimizer"] = self.optimizer.state_dict()
        save["last_batch_iteration"] = self.last_batch_iteration 
        save["cycle"] = self.cycle 
        return save

class one_cycle():
    def __init__(self, optimizer, batch_size, total_images, stepsize, max_lr, base_lr, max_mom, base_mom, epochs):
        self.optimizer = optimizer
        self.max_lr = max_lr
        self.base_lr = base_lr
        self.max_mom = max_mom
        self.base_mom = base_mom
        self.onecycle = False
        self.batch_size = batch_size 
        self.total_images = total_images
        self.total_ep = epochs
        self.last_batch_iteration = 0
        self.iterations_per_batch = math.ceil(self.total_images/self.batch_size)
        self.stepsize = stepsize * self.iterations_per_batch
    
    def get_triangular_lr(self):
        cycle = np.floor(1 + self.last_batch_iteration/(2  * self.stepsize))
        x = np.abs(self.last_batch_iteration/self.stepsize - 2 * cycle + 1)
        lr = self.base_lr + (self.max_lr - self.base_lr) * np.maximum(0, (1-x))
        return lr

    def get_inverted_triangular_mom(self):
        cycle = np.floor(1 + self.last_batch_iteration/(2  * self.stepsize))
        x = np.abs(self.last_batch_iteration/self.stepsize - 2 * cycle + 1)
        mom = self.max_mom + (self.base_mom - self.max_mom) * np.maximum(0, (1-x))
        return mom
    
    def fix_lr(self, *argv):
        new_lr = self.get_triangular_lr()
        new_mom = self.get_inverted_triangular_mom()
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = new_lr
            param_group["momentum"] = new_mom
        if (self.last_batch_iteration==2*self.stepsize):
            self.last_batch_iteration = 0
            self.stepsize = self.total_ep - (self.stepsize/self.iterations_per_batch)*2
            self.max_lr = 0
            self.base_mom = self.max_mom
        
        self.last_batch_iteration += 1
        return new_lr

    def load_state_dict(self, save):
        self.optimizer.load_state_dict(save["optimizer"])
        self.last_batch_iteration = save["last_batch_iteration"]
        self.stepsize = save["stepsize"]
    
    def state_dict(self):
        save = {}
        save["optimizer"] = self.optimizer.state_dict()
        save["last_batch_iteration"] = self.last_batch_iteration
        save["stepsize"] = self.stepsize
        return save


class cyclic_lr():
    def __init__(self, optimizer, batch_size, total_images, stepsize, max_lr, base_lr):
        self.optimizer = optimizer
        self.max_lr = max_lr
        self.base_lr = base_lr
        self.batch_size = batch_size 
        self.total_images = total_images
        self.last_batch_iteration = 0
        self.iterations_per_epoch = math.floor(self.total_images/self.batch_size)
        self.stepsize = stepsize * self.iterations_per_epoch
    
    def get_triangular_lr(self):
        cycle = np.floor(1 + self.last_batch_iteration/(2  * self.stepsize))
        x = np.abs(self.last_batch_iteration/self.stepsize - 2 * cycle + 1)
        lr = self.base_lr + (self.max_lr - self.base_lr) * np.maximum(0, (1-x))
        self.last_batch_iteration += 1
        return lr
    
    def fix_lr(self, *argv):
        new_lr = self.get_triangular_lr()
        for param_group in self.optimizer.param_groups:
            param_group["lr"] = new_lr
        return new_lr

    def state_dict(self):
        return {key: value for key, value in self.__dict__.items() if key not in {'optimizer'}}

    def load_state_dict(self, state_dict):
        self.__dict__.update(state_dict)



class ReduceLRonPlateau():
    def __init__(self, optimizer, batch_size, total_images, mode, factor, patience,verbose, threshold, 
                 threshold_mode, cooldown, min_lr, eps):
        self.factor = factor
        self.optimizer = optimizer

        if isinstance(min_lr, list) or isinstance(min_lr, tuple):
            if len(min_lr) != len(optimizer.param_groups):
                raise ValueError("expected {} min_lrs, got {}".format(len(optimizer.param_groups), len(min_lr)))
            self.min_lrs = list(min_lr)
        else:
            self.min_lrs = [min_lr] * len(optimizer.param_groups)

        self.patience = patience
        self.verbose = verbose
        self.cooldown = cooldown
        self.cooldown_counter = 0
        self.mode = mode
        self.threshold = threshold
        self.threshold_mode = threshold_mode
        self.best = None
        self.num_bad_epochs = None
        self.mode_worse = None  # the worse value for the chosen mode
        self.is_better = None
        self.eps = eps
        self.last_epoch = -1
        self.batch_size = batch_size 
        self.total_images = total_images
        self.last_batch_iteration = -1
        self.iterations_per_batch = math.ceil(self.total_images/self.batch_size)
        self._init_is_better(mode=mode, threshold=threshold, threshold_mode=threshold_mode)
        self._reset()

    def _reset(self):
        """Resets num_bad_epochs counter and cooldown counter."""
        self.best = self.mode_worse
        self.cooldown_counter = 0
        self.num_bad_epochs = 0

    def fix_lr(self, metrics, epoch=None):
        current = metrics
        self.last_batch_iteration += 1
        if (self.last_batch_iteration!=self.iterations_per_batch):
            return self.optimizer.param_groups[0]['lr']
        else:
            self.last_batch_iteration = 0
            if epoch is None:
                epoch = self.last_epoch = self.last_epoch + 1
            self.last_epoch = epoch

            if self.is_better(current, self.best):
                self.best = current
                self.num_bad_epochs = 0
            else:
                self.num_bad_epochs += 1

            if self.in_cooldown:
                self.cooldown_counter -= 1
                self.num_bad_epochs = 0  # ignore any bad epochs in cooldown

            if self.num_bad_epochs > self.patience:
                self._reduce_lr(epoch)
                self.cooldown_counter = self.cooldown
                self.num_bad_epochs = 0
            return self.optimizer.param_groups[0]['lr']

    def _reduce_lr(self, epoch):
        for i, param_group in enumerate(self.optimizer.param_groups):
            old_lr = float(param_group['lr'])
            new_lr = max(old_lr * self.factor, self.min_lrs[i])
            if old_lr - new_lr > self.eps:
                param_group['lr'] = new_lr
                if self.verbose:
                    print('Epoch {:5d}: reducing learning rate'
                          ' of group {} to {:.4e}.'.format(epoch, i, new_lr))

    @property
    def in_cooldown(self):
        return self.cooldown_counter > 0

    def _cmp(self, mode, threshold_mode, threshold, a, best):
        if mode == 'min' and threshold_mode == 'rel':
            rel_epsilon = 1. - threshold
            return a < best * rel_epsilon

        elif mode == 'min' and threshold_mode == 'abs':
            return a < best - threshold

        elif mode == 'max' and threshold_mode == 'rel':
            rel_epsilon = threshold + 1.
            return a > best * rel_epsilon

        else:  # mode == 'max' and epsilon_mode == 'abs':
            return a > best + threshold

    def _init_is_better(self, mode, threshold, threshold_mode):
        if mode not in {'min', 'max'}:
            raise ValueError('mode ' + mode + ' is unknown!')
        if threshold_mode not in {'rel', 'abs'}:
            raise ValueError('threshold mode ' + threshold_mode + ' is unknown!')

        if mode == 'min':
            self.mode_worse = float('inf')
        else:  # mode == 'max':
            self.mode_worse = (-float('inf'))

        self.is_better = partial(self._cmp, mode, threshold_mode, threshold)

    def state_dict(self):
        return {key: value for key, value in self.__dict__.items() if key not in {'optimizer', 'is_better'}}

    def load_state_dict(self, state_dict):
        self.__dict__.update(state_dict)
        self._init_is_better(mode=self.mode, threshold=self.threshold, threshold_mode=self.threshold_mode)


class StepLR():
    def __init__(self, optimizer, batch_size, total_images, step_size, gamma, last_epoch):
        self.optimizer = optimizer
        if last_epoch == -1:
            for group in optimizer.param_groups:
                group.setdefault('initial_lr', group['lr'])
        else:
            for i, group in enumerate(optimizer.param_groups):
                if 'initial_lr' not in group:
                    raise KeyError("param 'initial_lr' is not specified "
                                   "in param_groups[{}] when resuming an optimizer".format(i))
        self.base_lrs = list(map(lambda group: group['initial_lr'], optimizer.param_groups))
        self.last_epoch = last_epoch
        self.step_size = step_size
        self.gamma = gamma
        self.batch_size = batch_size 
        self.total_images = total_images
        self.iterations_per_batch = math.ceil(self.total_images/self.batch_size)
        self.last_batch_iteration = self.iterations_per_batch - 1
        self.fix_lr(last_epoch + 1)
        self.last_batch_iteration = -1

    def get_lr(self):
        return [base_lr * self.gamma ** (self.last_epoch // self.step_size)
                for base_lr in self.base_lrs]

    def fix_lr(self, metrics, epoch=None):
        self.last_batch_iteration += 1
        if (self.last_batch_iteration!=self.iterations_per_batch):
            return self.optimizer.param_groups[0]['lr']
        self.last_batch_iteration = 0
        if epoch is None:
            epoch = self.last_epoch + 1
        self.last_epoch = epoch
        for param_group, lr in zip(self.optimizer.param_groups, self.get_lr()):
            param_group['lr'] = lr
        return self.optimizer.param_groups[0]['lr']

    def state_dict(self):
        return {key: value for key, value in self.__dict__.items() if key not in {'optimizer'}}

    def load_state_dict(self, state_dict):
        self.__dict__.update(state_dict)


import torch
import numpy as np
import torch.nn.functional as F
from sklearn.metrics import fbeta_score, f1_score

def get_metric(opt, outputs, targets):
    """ Metric to evaluate

    outputs = [network outputs]
    targets = [network targets]
    """
    if opt.metric == "accuracy":
        outputs = F.softmax(outputs, dim =1)
        _, pred = outputs.topk(1)
        accuracy = float(torch.sum(pred == targets).cpu().numpy())/pred.shape[0]
        
        if opt.metric_n:
            _, pred = outputs.topk(opt.metric_n)
            images, _ = targets.size()
            targets = targets.expand(images, opt.metric_n)
            accuracy_n = float(torch.sum(pred == targets).cpu().numpy())/pred.shape[0]
            
        return accuracy, accuracy_n
        
    elif opt.metric == "logloss":
        return float(F.cross_entropy(outputs, targets.view(-1), weight=opt.class_weight).cpu().numpy()), _

    elif opt.metric == "mean_f2_score":
        out = torch.sigmoid(outputs)
        out = out.detach().cpu().numpy()
        #print(targets)
        #print(out)
        out = out>0.5
        return fbeta_score(targets, out, average='weighted', beta=2), 0
    
    elif opt.metric == "f1_macro":
        out = torch.sigmoid(outputs)
        out = out.detach().cpu().numpy()
        out = out > 0.5
        return f1_score(targets.cpu().numpy(), out, average="macro"), 0

    else:
        raise NotImplementedError



import torch
import numpy as np 
import skimage
from tqdm import tqdm 
import os 
import pickle


from albumentations import (HorizontalFlip, VerticalFlip, CLAHE, RandomContrast, RandomGamma, RandomBrightness, MedianBlur, IAASharpen, IAAEmboss, HueSaturationValue, ToGray, OneOf, RGBShift, ChannelShuffle, Compose)

from .img_transforms import MyResizer, Normalizer 



class ListDataset():
    def __init__(self,
                 opt,
                 train=True):
        self.opt = opt
        self.train = train
        self.multiclass = self.opt.multiclass
        #self.aspect_loc = self.opt.aspect_loc
        if self.multiclass:
            from sklearn.preprocessing import MultiLabelBinarizer
            self.one_hot = MultiLabelBinarizer(classes=np.arange(self.opt.num_classes))
        if self.train:
            self.list_file = self.opt.train_data_loc
        else:
            self.list_file = self.opt.val_data_loc
            
        self.data_loc = self.opt.data_loc
        
        
        if self.train:
            self.transforms = ListDataset.train_transformations(self.opt)
        self.images, self.labels = [], []

        with open(self.list_file) as f:
            lines = f.readlines()
            self.num_samples = len(lines)

        for line in lines:
            splited = line.strip().rsplit(" ")
            self.images.append(splited[0])
            if self.multiclass:
                self.labels.append([int(i) for i in splited[1:]])
            else:
                self.labels.append(torch.LongTensor([int(splited[1])]))
        print("images:", len(self.images))
        if self.train:
            self.aspect_ratios = self.load_aspect_ratio(self.opt.aspect_loc_train)
        else:
            self.aspect_ratios = self.load_aspect_ratio(self.opt.aspect_loc_valid)
        
        if self.opt.normalize:
            self.normalizer = Normalizer(self.opt.mean, self.opt.std)
        if self.opt.resize:
            self.resize = MyResizer(min_side = self.opt.min_size)

    def __getitem__(self, idx):
        img_loc = self.images[idx]
        label = self.labels[idx]
        img, scale = self.get_image(img_loc)
        if self.multiclass:
            label = torch.FloatTensor(self.one_hot.fit_transform([label])).view(-1)
        else:
            label = label.long()
        output = {"img": img, "scale": scale, "label": label}
        return output

    def __len__(self):
        return len(self.images)

    def get_image(self, img_loc):
        img = self.load_image(img_loc)
        img = np.asarray(img)
        if self.train:
            img = self.transforms(image = img)
        else:
            img = {"image": img}
        if self.opt.normalize:
            img = self.normalizer(img)
        if self.opt.resize:
            img = self.resize(img)
            
        return img["image"], img["scale"]

    def inverse_image(self):
        pass ## Given torch output tensor output of a dataloader, this should reverse engineer the options and output another image.
    
    def load_image(self, loc):
        path = self.data_loc + loc
        img = skimage.io.imread(path)
        if len(img.shape) == 2:
            img = skimage.color.gray2rgb(img)
        return img.astype(np.float32)/255.0
            
    
    @staticmethod
    def train_transformations(opt):
        """Albugumentations 
        
        """
        img_transforms = []
        if opt.hflip:
            img_transforms.append(HorizontalFlip(p=0.5))
        if opt.vflip:
            img_transforms.append(VerticalFlip(p=0.5))
        if opt.color_transforms:
            color_transforms = [
                                RGBShift(p=0.5),
                                ChannelShuffle(p=0.5),
                                CLAHE(p=0.5),
                                RandomContrast(p=0.5),
                                RandomGamma(p=0.5),
                                RandomBrightness(p=0.5),
                                MedianBlur(p=0.1),
                                IAASharpen(),
                                IAAEmboss()]
            img_transforms.extend(color_transforms)
            img_transforms.append(HueSaturationValue(p=0.3))
        if opt.togray:
            img_transforms.append(ToGray(p=0.05))
        return Compose(img_transforms)
    
#     @staticmethod
#     def test_transformations_new(opt):
#         img_transforms = []
#         if opt.normalize:
#             img_transforms.append(Normalizer(opt.mean, opt.std))
#         if opt.resize:
#             img_transforms.append(MyResizer(min_side=opt.min_size))
#         return Compose(img_transforms)

    def load_aspect_ratio(self, save_loc="img_attributes.pth"):
        import pickle
        if os.path.isfile(save_loc):
            with open(save_loc, 'rb') as handle:
                aspect_ratios = pickle.load(handle)
        else:
            aspect_ratios = list()
            for i in tqdm(range(self.num_samples)):
                img = self.load_image(self.images[i])
                aspect_ratios.append(float(img.shape[1])/float(img.shape[0]))
            with open(save_loc, 'wb') as handle:
                pickle.dump(aspect_ratios, handle, protocol=pickle.HIGHEST_PROTOCOL)
        assert len(aspect_ratios) == self.num_samples
        #print(aspect_ratios)
        return aspect_ratios
    
    def image_aspect_ratio(self, idx):
        return self.aspect_ratios[idx]      



import torch
import torch.nn as nn

import math
import sys
import numpy as np

from ..anchors.encoder import process_yolo_output, encode_targets
from ..vis_lib.visdom_vis import VisdomLinePlotter
from ..anchors.anchors_generator import generate_anchors


class Yolov3_loss(nn.Module):
    def __init__(self, num_classes = 80, use_gpu=True, use_visdom=True, exp_name="yolo"):
        super(Yolov3_loss, self).__init__()
        self.num_classes=num_classes
        self.use_gpu = use_gpu
        self.use_visdom = use_visdom 
        
        if self.use_visdom:
            print("Using visdom: run visdom.server somewhere")
            self.plot = VisdomLinePlotter(env_name = exp_name)
        self.iterations = 1

    
    def forward(self, cls_preds, cls_targets, scales, stride, nW, nH, pyramids=3):
        #cls_preds = cls_preds.cpu()
        #cls_targets = cls_targets.cpu()
        #print(stride, nW, nH, scales)
        anchors = generate_anchors(stride, nW, nH, scales)
        if self.use_gpu:
            anchors = anchors.cuda()
        outs, scores = process_yolo_output(cls_preds, stride, scales, anchors.clone(), nW, nH, self.num_classes)
        coord_mask, conf_mask, cls_mask, tcoord, tconf, tcls, nRecall, nRecall75, nGT = encode_targets(outs.cpu().clone(),
                                                                                                    cls_targets.cpu().detach(), 
                                                                                                    scales,
                                                                                                    anchors.cpu().clone(),
                                                                                                    nW, 
                                                                                                    nH, 
                                                                                                    self.num_classes,
                                                                                                    stride)
        cls = scores[:,:, 5:].clone()
        coord = scores[:,:, :4].clone()
        conf = scores[:,:,  4].clone()
        
        #print("conf", conf.cpu().unique())
        #print("tconf", tconf.sum())
        
        if self.use_gpu:
            coord_mask = coord_mask.cuda()
            conf_mask = conf_mask.cuda()
            cls_mask = cls_mask.cuda()
            tcoord = tcoord.cuda()
            tconf = tconf.cuda()
            tcls = tcls.cuda()
        nProposals = int((conf > 0.25).float().sum())
        
        ## cls score 
        #cls = cls[(cls_mask == 1)]
        #tcls = tcls[(cls_mask == 1)].long().view(-1) - 1 ## 0-79 rathere than 1-80
        #loss_cls = torch.nn.CrossEntropyLoss(size_average=False)(cls, tcls) if cls.size(0) > 0 else 0
        summed = tcls.sum()
        loss_cls = torch.nn.BCELoss(size_average=False)(cls[(tcls ==1)], cls_mask[(tcls==1)] )/(summed)
        
        ## coord_loss
        #coord_mask = coord_mask.unsqueeze(2)
        loss_coord = 4*(torch.nn.MSELoss(size_average=False)(coord[(tcls==1)], tcoord[(tcls==1)])/(summed))
        
        ## conf_loss 
        #loss_conf = torch.nn.BCELoss(size_average=False)(conf[(conf_mask ==1)], tconf[(conf_mask == 1)])/ (summed)
        loss_conf = torch.nn.MSELoss(size_average=False)(conf*conf_mask, tconf*conf_mask)/(summed)
        total_loss = loss_conf + loss_coord + loss_cls
        print("Iter: {}, nProposals: {}, nRecall: {}, nRecall75: {}, nGT: {}, loss_conf: {}, loss_coord: {}, loss_cls: {}, total_loss: {}".format(int(self.iterations/pyramids), nProposals, nRecall, nRecall75, nGT, loss_conf, loss_coord, loss_cls, total_loss))
        
        if self.iterations%10 == 0: 
            if self.use_visdom:
                self.plot.plot("loss_coord", "training", self.iterations, np.array(float(loss_coord)))
                self.plot.plot("loss_conf", "training", self.iterations, np.array(float(loss_conf)))
                self.plot.plot("loss_cls", "training", self.iterations, np.array(float(loss_cls)))

        self.iterations +=1
        return loss_conf, loss_coord, loss_cls


import torch
from fractal.losses.yolov3_loss import Yolov3_loss

def get_loss(loss_type):
    if loss_type == "yolov3_loss":
        print("Using yolov3 loss model")
        return Yolov3_loss
    else:
        raise NotImplementedError("We haven't implemented these Loss function yet. Please raise a pull request if it is useful")


import torch
import torchvision
import math
import torch.nn as nn

def resnet18(num_classes, pretrained=True, imbalance=False, prior=0.01):
    model = torchvision.models.resnet18(pretrained=pretrained)
    num_fltrs = model.fc.in_features
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    model.fc = torch.nn.Linear(num_fltrs, num_classes)
    if imbalance:
        model.fc.weight.data.fill_(0)
        model.fc.bias.data.fill_(-math.log((1.0-prior)/prior))
    return model
def resnet50(num_classes, pretrained=True, imbalance=False, prior=0.01):
    model = torchvision.models.resnet50(pretrained=pretrained)
    num_fltrs = model.fc.in_features
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    model.fc = torch.nn.Linear(num_fltrs, num_classes)
    if imbalance:
        model.fc.weight.data.fill_(0)
        model.fc.bias.data.fill_(-math.log((1.0-prior)/prior))
    return model

def resnet101(num_classes, pretrained=True, imbalance=False, prior=0.01):
    model = torchvision.models.resnet101(pretrained=pretrained)
    num_fltrs = model.fc.in_features
    model.avgpool = nn.AdaptiveAvgPool2d(1)
    model.fc = torch.nn.Linear(num_fltrs, num_classes)
    if imbalance:
        model.fc.weight.data.fill_(0)
        model.fc.bias.data.fill_(-math.log((1.0-prior)/prior))
    return model



import torch
import torchvision.transforms as transforms 
from PIL import Image

import numpy as np 

from ..anchors.utils import clamper, bbox_iou

def get_grid_xy(img_size, stride):
    """ Given img_size and stide, it extracts the xy locations (top, left) of all the image patches
    
    img_size: tuple: Size of the image (h, w)
    stride: int: stride to move along x and y axis 
    
    Returns:
    x_y_offset: [N, 2]: A image tensor
    """
    h = img_size[0]//stride
    w = img_size[1]//stride
    grid_x = np.arange(w)
    grid_y = np.arange(h)
    a, b = np.meshgrid(grid_x, grid_y)
    x_offset = torch.FloatTensor(a).view(-1, 1)
    y_offset = torch.FloatTensor(b).view(-1, 1)
    x_y_offset = torch.cat((x_offset, y_offset), 1)
    return x_y_offset

def patch_loc(stride, patch_size, img_size, pad=True):
    """ Extracts the patch locations from the image by using padding 
    
    stride: int: Stride to move along x and y axis of the image : Future scope is to move 3D 
    patch_size: int: Size of the image patch we need 
    img_size: tuple: Size of the image patch we need 
    pad: boolean: padding to convert image to extract all parts of the image with equal size
    
    Returns: 
    xywh: Tensor: [N, 4] Locations of the image patches (ctr_x, ctr_y, w, h)
    img_size: tuple: The transformed image size. use it to pad the image 
    """
    h, w = img_size 
    if pad:
        h_ = patch_size - (h%stride)
        w_ = patch_size - (w%stride)
        img_size = (h+h_, w+w_)
    x_y_offset = get_grid_xy(img_size, stride)
    xy = x_y_offset * stride 
    xy[:, 0] = xy[:, 0] + (patch_size/2)
    xy[:, 1] = xy[:, 1] + (patch_size/2)
    wh = torch.ones(xy.shape) * patch_size 
    xywh = torch.cat([xy, wh], 1)
    return xywh, img_size 

def read_img(loc, tensor=True):
    """ Read a image into python environment
    """
    img = Image.open(loc)
    if tensor:
        img = transforms.ToTensor()(img)
    return img 

def pil_2_tensor(pil_img):
    """convert a pil image to tensor
    """
    img = transforms.ToTensor()(pil_img)
    return img


def pad_img(img_pil, new_dim):
    """Pad the image so that the output image will have the dimensions of new dim. Always use "constant" mode only to pass
    """
    w, h = img_pil.size
    h_, w_ = new_dim 
    h_pad = h_ - h
    w_pad = w_ - w 
    padded_img = transforms.functional.pad(img_pil, padding=(w_pad//2, h_pad//2), padding_mode="constant")
    return padded_img 
    

def get_img_patches(img_tensor, grid_dim):
    """Gets image patches tensors according to the grid
    
    Inputs:
    --------
    img_tensor: (C, H, W) - Image tensor
    grid_dim: [y1, x1, y2, x2] [N, 4] - N patchces xyxy co-ordinates 

    Outputs:
    --------
    Lists of Tensors (Each tensor represents the part of a images)
    """
    return [img_tensor[:, int(i):int(k), int(j): int(l)] for i, j, k, l in grid_dim]

def tensors_2_pil(img_tensors_list):
    """ Converts a list of tensors to PIL images 
    
    img_tensors_list: List of tensors which are converted to images
    """
    pills = transforms.ToPILImage()
    pil_list = [pills(img_tensors_list[i]) for i in range(len(img_tensors_list))]
    return pil_list

def bbox_extractor(img_grid, bbox, labels, patch_size, iou_required=0.5):
    """ Assign bboxes to respective grids and change their co-ordinates with respect to that bbox 
    
    Inputs:
    -------
    img_grid: [N, 4]: yxyx format. The coord loc of grid for each image: torch.FloatTensor 
    bbox: [N, 4]: yxyx format. The coord loc of bboxes of a image: torch.FloatTensor
    
    Outputs:
    --------
    A dict: with each key has image patch index and its key as bboxes assoiciated with it.
    
    bboxes: [N, 6]: [y1, x1, y2, x2, label , iou]
    """
    bbox_xy = bbox[:, [1,0, 3, 2]].clone()
    ious_area = bbox_iou(img_grid.numpy(), bbox.numpy(), intersection=False) # numpy is bloating man, Change this quickly
    grids, boxes = ious_area.shape
    bbox_area = (bbox[:, 2] - bbox[:, 0]) * (bbox[:, 3] - bbox[:, 1])
    final_ious = ious_area/bbox_area.repeat(grids).view(-1, boxes)
    
    patch_index = {} # key as patch number and values as boxes assoiciated with it 
    for i in range(img_grid.shape[0]):
        mask = (final_ious[i, :] > iou_required)
        boxes_ = bbox[mask] #yxyx
        labels_ = labels[mask]
        ious_ = final_ious[i, :][mask]
        patch_ = img_grid[i, :]
        ## convertin yxyx to yxhw ---> yx is (top left)
        
        
        if len(boxes_) > 0:
            boxes_[:, 2] = boxes_[:, 2] - boxes_[:, 0] # h
            boxes_[:, 3] = boxes_[:, 3] - boxes_[:, 1] # w 
            #boxes_ = boxes_[:, [0, 1, 3, 2]] # y1x1wh 
            boxes_[:, :2] = boxes_[:, :2] - patch_[:2]
            boxes_[:, 2] = boxes_[:, 0] + boxes_[:, 2] # y2
            boxes_[:, 3] = boxes_[:, 1] + boxes_[:, 3]
            #boxes_[:, 2:] = boxes_[:, 2:] - patch_[2:] + patch_size 
            boxes_ = clamper(boxes_[:, [1, 0, 3, 2]], (patch_size, patch_size)) #converting yxyx format to xyxy
            boxes_ = torch.cat([boxes_, labels_.float().view(-1, 1), ious_.view(-1, 1)], 1)
        patch_index[i] = boxes_
    return patch_index


import torch
import numpy as np
import matplotlib.pyplot as plt
from matplotlib import patches
import colorsys

from PIL import Image, ImageFont, ImageDraw
from ..anchors.utils import xywh_xyxy

## printing a few bounding boxes

def draw_grid(image, bbox, outline="white", input_format="xyxy"):
    """draws rectangles on the image given by bbox 
    
    image: PIL image 
    bbox: numpy with each box representing the format defined by "format"
    outline: color of the bbox 
    input_format: "xxyy" or "xyhw": use one of this 
    
    """
    draw = ImageDraw.Draw(image)
    if input_format == "xywh":
        bbox = xywh_xyxy(bbox)
    
    if isinstance(bbox, torch.Tensor):
        bbox = bbox.numpy()
        
    for i in bbox:
        x0, y0, x1, y1 = i
        draw.rectangle([x0, y0, x1, y1], outline=outline)
    return image

def draw_centre(image, bbox, outline="white", input_format="xyxy"):
    r = (image.size[0]+image.size[1])//600
    print(r)
    draw = ImageDraw.Draw(image)
    
    if input_format == "xywh":
        bbox = xywh_xyxy(bbox)
    
    if isinstance(bbox, torch.Tensor):
        bbox = bbox.numpy()
    
    for i in bbox:
        x0, y0, x1, y1 = i
        x_ctr = (x1+x0)/2
        y_ctr = (y1+y0)/2
        draw.ellipse([x_ctr-r, y_ctr-r, x_ctr+r, y_ctr+r], outline=outline)
    return image


def vis_bbox(image, bbox, color="red"):
    """ Randomly plots every 20th anchor box

    image = np.zeros((800, 600, 3)).astype(np.uint8)
    bbox = anchor[index_inside, :]
    """
    print(image.shape)
    fig = plt.figure(figsize=(416//52, 416//52))
    ax = fig.subplots(1)
    ax.imshow(image)
    for m, i in enumerate(bbox):
        xmin, ymin, xmax, ymax = i
        height = ymax - ymin
        width = xmax - xmin

        if np.random.randint(0, 20) ==0:
            rect = patches.Rectangle([xmin, ymin], width, height, color = color, fill=False)
            ax.add_patch(rect)
    plt.show()


def vis_bbox_inference(image, bbox, color="red"):
    """ Plots all the anchor boxes. No randomness involved
    image = np.zeros((800, 600, 3)).astype(np.uint8)
    bbox = anchor[index_inside, :]
    """
    print(image.shape)
    fig = plt.figure(figsize=(416//52, 416//52))
    ax = fig.subplots(1)
    ax.imshow(image)
    for m, i in enumerate(bbox):
        xmin, ymin, xmax, ymax = i
        height = ymax - ymin
        width = xmax - xmin
        rect = patches.Rectangle([xmin, ymin], width, height, color = color, fill=False)
        ax.add_patch(rect)
    plt.show()


def save_img_with_labels(image, bboxes, prob_scores, labels, labels_mapping, save_loc="pred_img.png",
                        font_loc="fractal/vis_lib/FiraMono-Medium.otf"):
    """saves the image to input location

    image: Input PIL image
    bboxes: bounding boxes of the image. ndarray [N, 4]
    prob_scores: prob_scores of the boxes. ndarray [N,]
    labels: labels for the box . ndarray [N,]
    labels_mapping: mapping of the labels dict example: {1:"person", 2: "box"}
    save_loc: location to save the image

    Returns:
    save the PIL image at given location if save_loc is not None, else outputs a PIL image with bboxes
    """
    hsv_tuples = [(x/len(labels_mapping), 1., 1.) for x in range(len(labels_mapping))]
    colors = list(map(lambda x: colorsys.hsv_to_rgb(*x), hsv_tuples))
    colors = list(map(lambda x: (int(x[0]*255), int(x[1]*255), int(x[2]*255)), colors))

    font = ImageFont.truetype(font=font_loc, size=np.floor(3e-2 * image.size[1]+0.5).astype("int32"))
    thickness = (image.size[0]+image.size[1]) // 300

    for i, c in enumerate(labels):
        predicted_class = labels_mapping[c]
        box = bboxes[i]
        score = prob_scores[i]

        label = '{} {:.2f}'.format(predicted_class, score)

        draw = ImageDraw.Draw(image)

        label_size = draw.textsize(label, font)
        
        xmin, ymin, xmax, ymax = box
        #ymin, xmin, ymax, xmax = box
        bottom = int(ymax)
        top = int(ymin)
        left = int(xmin)
        right = int(xmax)

        if top - label_size[1] >=0:
            text_origin = np.array([left, top-label_size[1]])
        else:
            text_origin = np.array([left, top+1])
        
        for i in range(thickness):
            draw.rectangle(
            [left+i, top+i, right-i, bottom-i], outline=colors[int(c)])

        draw.rectangle([tuple(text_origin), tuple(text_origin+label_size)], fill=colors[int(c)])
        draw.text(text_origin, label, fill=(0, 0, 0), font=font)
        del draw

    if save_loc is not None:
        image.save(save_loc, quality=90)
    else:
        return image


import torch
import torch.nn as nn
import torch.nn.init as init
import math

from ..common_utils.img_utils import write_results


def initialize_weights(net):
    for m in net.modules():
        if isinstance(m, nn.Conv2d):
            init.normal(m.weight, mean=0, std=0.01)
            if m.bias is not None:
                init.constant(m.bias, 0)
        elif isinstance(m, nn.BatchNorm2d):
            m.weight.data.fill_(1)
            m.bias.data.zero_()
    return net

class ConvBlock(nn.Module):
    def __init__(self, inplanes, planes, kernel_size, stride=1, pad=1, bias=False):
        super(ConvBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_channels=inplanes,
                             out_channels = planes,
                             kernel_size = kernel_size,
                             bias = bias,
                             stride = stride,
                             padding = pad)
        self.bn1 = nn.BatchNorm2d(planes)
        self.lrelu = nn.LeakyReLU(negative_slope=0.1, inplace=True)

    def forward(self, x):
        out = self.conv1(x)
        out = self.bn1(out)
        out = self.lrelu(out)
        return out
    
    
class ResBlock(nn.Module):
    def __init__(self, in_channels, reduce):
        super(ResBlock, self).__init__()
        self.conv1 = ConvBlock(in_channels, reduce, 1, 1, 0)
        self.conv2 = ConvBlock(reduce, in_channels, 3, 1, 1)
    
    def forward(self, x):
        out = self.conv1(x)
        out = self.conv2(out)
        return out+x
        

class Darknet53(nn.Module):
    def __init__(self):
        super(Darknet53, self).__init__()
        self.conv1 = ConvBlock(3, 32, 3, 1, 1) # 256 * 256
        self.conv2 = ConvBlock(32, 64, 3, 2, 1) # 128 * 128
        self.resblock1 = self._make_layers(64, 32, 1)
        self.conv3 = ConvBlock(64, 128, 3, 2, 1)
        self.resblock2 = self._make_layers(128, 64, 2)
        self.conv4 = ConvBlock(128, 256, 3, 2, 1)
        self.resblock3 = self._make_layers(256, 128, 8)
        self.conv5 = ConvBlock(256, 512, 3, 2, 1)
        self.resblock4 = self._make_layers(512, 256, 8)
        self.conv6 = ConvBlock(512, 1024, 3, 2, 1)
        self.resblock5 = self._make_layers(1024, 512, 4)

    
    def _make_layers(self, in_channels, reduce, blocks):
        layer = []
        for _ in range(blocks):
            layer.append(ResBlock(in_channels, reduce))
        return nn.Sequential(*layer)
    
    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.resblock1(x)
        x = self.conv3(x)
        x = self.resblock2(x)
        x = self.conv4(x)
        out1 = self.resblock3(x)
        x = self.conv5(out1)
        out2 = self.resblock4(x)
        x = self.conv6(out2)
        out3 = self.resblock5(x)
        return out1, out2, out3
    

class Yolov3_classifier(nn.Module):
    def __init__(self, num_classes = 211):
        super(Yolov3_classifier, self).__init__()
        self.num_classes = num_classes
        self.backend = Darknet53()
        self.avgpool = nn.AvgPool2d(11, stride=1)
        self.classifier = nn.Linear(46080, self.num_classes)
    
    def forward(self, x):
        _, _, x = self.backend(x)
        x = self.avgpool(x)
        x = x.view(x.size(0), -1)
        x = self.classifier(x)
        return x 
    
    
    
class Darknet_Dbox(nn.Module):
    def __init__(self, num_classes=80, anchors=9):
        super(Darknet_Dbox, self).__init__()
        self.num_classes = num_classes
        self.anchors = anchors
        self.output_dim = int((self.num_classes+5) * (self.anchors/3))
        self.backend = Darknet53()
        self.conv_block_1 = nn.Sequential(*[
            ConvBlock(1024, 512, 1, 1, 0),
            ConvBlock(512, 1024, 3, 1, 1),
            ConvBlock(1024, 512, 1, 1, 0),
            
        ]) 
        self.yolo1_0 = nn.Sequential(*[
            ConvBlock(512, 1024, 3, 1, 1),
            ConvBlock(1024, 512, 1, 1, 0)
        ]) 
        self.yolo1_1 = nn.Sequential(*[
            ConvBlock(512, 1024, 3, 1, 1),
            nn.Conv2d(1024, self.output_dim, 1, 1, 0)
        ])# mask of [116,90,  156,198,  373,326], classes=80, num=9 jitter=.3, ignore_thresh=.7, thruth_thresh=1, random=1
        
        self.upsample1 = nn.Sequential(*[
            ConvBlock(512, 256, 1, 1, 0),
            nn.Upsample(scale_factor = 2, mode = "nearest"),
        ])
        
        self.conv_block_2 = nn.Sequential(*[
            ConvBlock(768, 256, 1, 1, 0),
            ConvBlock(256, 512, 3, 1, 1),
            ConvBlock(512, 256, 1, 1, 0),            
        ])
        
        self.yolo2_0 = nn.Sequential(*[
            ConvBlock(256, 512, 3, 1, 1),
            ConvBlock(512, 256, 1, 1, 0)
        ])
        self.yolo2_1 = nn.Sequential(*[
            ConvBlock(256, 512, 3, 1, 1),
            nn.Conv2d(512, self.output_dim, 1, 1, 0)
        ])# mask of [ 30,61,  62,45,  59,119], classes=80, num=9 jitter=.3, ignore_thresh=.7, thruth_thresh=1, random=
        
        
        self.upsample2 = nn.Sequential(*[
            ConvBlock(256, 128, 1, 1, 0),
            nn.Upsample(scale_factor = 2, mode = "nearest"),
        ])
        
        self.conv_block_3 = nn.Sequential(*[
            ConvBlock(384, 128, 1, 1, 0),
            ConvBlock(128, 256, 3, 1, 1),
            ConvBlock(256, 128, 1, 1, 0),            
        ])
        self.yolo3_0 = nn.Sequential(*[
            ConvBlock(128, 256, 3, 1, 1),
            ConvBlock(256, 128, 1, 1, 0)
        ])
        self.yolo3_1 = nn.Sequential(*[
            ConvBlock(128, 256, 3, 1, 1),
            nn.Conv2d(256, self.output_dim, 1, 1, 0)
        ])# mask of [ 10,13,  16,30,  33,23], classes=80, num=9 jitter=.3, ignore_thresh=.7, thruth_thresh=1, random=1
        
        self.__init_this_shit(self.yolo1_1)
        self.__init_this_shit(self.yolo2_1)
        self.__init_this_shit(self.yolo3_1)
        
        
    
    def __init_this_shit(self, x, val =0.2):
        for num, m in enumerate(x):
            #print(num, m)
            if isinstance(m, nn.Conv2d):
                torch.nn.init.normal_(m.weight, mean=-0.019, std = 0.05)
                if m.bias is not None:                
                    if num == 1:
                        #print("Bias is using this formula")
                        bias = -1 * math.log((1-val)/val)
                        print(bias)
                        torch.nn.init.constant_(m.bias, -0.05) #mean=-0.5, std=0.12)
                    else:
                        torch.nn.init.constant_(m.bias, 0)
        
    
    def forward(self, x):
        out1, out2, out3 = self.backend(x)
        x = self.conv_block_1(out3)
        x = self.yolo1_0(x)
        yolo1 = self.yolo1_1(x)
        out = self.upsample1(x)
        x = torch.cat((out, out2), 1) #512+256 #
        x = self.conv_block_2(x)
        x = self.yolo2_0(x)
        yolo2 = self.yolo2_1(x)
        out = self.upsample2(x)
        x = torch.cat((out, out1), 1) # 256+128
        x = self.conv_block_3(x)
        x = self.yolo3_0(x)
        yolo3 = self.yolo3_1(x)
        return yolo1, yolo2, yolo3

class Darknet_yolo(nn.Module):
    def __init__(self, num_classes=80, anchors=9, scale=3):
        super(Darknet_yolo, self).__init__()
        self.dpn = Darknet_Dbox(num_classes, anchors)
        self.num_classes = num_classes 
        self.anchors = anchors
        self.scale = scale
        
    def forward(self, x, inference=True):
        if inference:
            batch_size = x.shape[0]
            out = self.dpn(x)
            anchors_per_scale = int(self.anchors/len(out))

            yolo1, yolo2, yolo3 = out
            yolo3_grid_size = int(yolo3.shape[-1])
            yolo2_grid_size = int(yolo2.shape[-1])
            yolo1_grid_size = int(yolo1.shape[-1])

            yolov3_ = yolo3.view(batch_size, (self.num_classes+5) * anchors_per_scale, yolo3_grid_size * yolo3_grid_size).transpose(1, 2).contiguous()
            yolov3_ = yolov3_.view(batch_size, yolo3_grid_size*yolo3_grid_size*anchors_per_scale, (self.num_classes+5))

            yolov2_ = yolo2.view(batch_size, (self.num_classes+5) * anchors_per_scale, yolo2_grid_size * yolo2_grid_size).transpose(1, 2).contiguous()
            yolov2_ = yolov2_.view(batch_size, yolo2_grid_size*yolo2_grid_size*anchors_per_scale, (self.num_classes+5))

            yolov1_ = yolo1.view(batch_size, (self.num_classes+5) * anchors_per_scale, yolo1_grid_size * yolo1_grid_size).transpose(1, 2).contiguous()
            yolov1_ = yolov1_.view(batch_size, yolo1_grid_size*yolo1_grid_size*anchors_per_scale, (self.num_classes+5))
            return torch.cat([yolov1_, yolov2_, yolov3_], 1)
            
        else:
            batch_size = x.shape[0]
            out = self.dpn(x)
            #anchors_per_scale = self.anchors/len(out)

            yolo1, yolo2, yolo3 = out

            if self.scale  <= 3:
                return [yolo1, yolo2, yolo3]
            else:
                raise ValueError("Only 3 scales are available")
    
    def predict(self, img_tensor, anchors, strides, nms_applied=True, conf=0.5, iou=0.4):
        """ Outputs a tensor, along with bbox_coord, cls_label and prob_score.
        Inputs:
        -------
        img_tensor: [Batch_size, C, W, H] 
        anchors: [Batch_size, n_anchors, (self.num_classes+5)] # batchsize is 1 
        strides: [n_anchors]
        
        Returns:
        Torch Tensor [N, 6] --> [x1, y1, x2, y2, label, prob_score]
        
        """
        self.eval()
        torch.set_grad_enabled(False)
        out = self(img_tensor)
        out[:, :, 0] = torch.sigmoid(out[:, :, 0])
        out[:, :, 1] = torch.sigmoid(out[:, :, 1])
        out[:, :, 4] = torch.sigmoid(out[:, :, 4])
        
        out[:, :, :2] += anchors[:, :, :2]
        out[:, :, 2:4] = torch.exp(out[:, :, 2:4]) * anchors[:, :, 2:]
        out[:, :, 5:5+self.num_classes] = torch.sigmoid(out[:, :, 5:5+self.num_classes])
        
        out[:, :, :4] *= strides.view(-1, 1).repeat(1, 4).unsqueeze(0)
        out = write_results(out.cpu(), conf, self.num_classes, nms_applied,  iou)
        torch.set_grad_enabled(True)
        self.train()
        return out
        
        
    def load_weights_manually(self, weight_loc):
        """ Loading weights manually to allow transfer learning
        """
        loaded_weights = torch.load(weight_loc)
        
        if "model" in loaded_weights.keys():
            print("Using pretrained weights....")
            loaded_weights = loaded_weights["model"]
            loaded_weights = {'dpn.'+k: v for k, v in loaded_weights.items()}
            self.load_state_dict(loaded_weights)
        else:
            actual_weights = self.dpn.state_dict()
            output_dim = loaded_weights[list(actual_weights.keys())[-1]].shape[0]
            print(output_dim, self.dpn.output_dim)
            if output_dim != self.dpn.output_dim:
                print("Transfer learning... Transfering all weights except the last layer")
                for name1 in actual_weights.keys():
                    if (name1 in loaded_weights.keys()):
                        if actual_weights[name1].shape == loaded_weights[name1].shape:
                            actual_weights[name1] = loaded_weights[name1]
                            print("transferred weights {} {}".format(name1, name1))
                        else:
                            print("Randomly initialized weights for: {} {}".format(name1, name1))
            else:
                actual_weights = loaded_weights
            self.dpn.load_state_dict(actual_weights)
        return self


""" Datagenerator to train the model

The list file looks like this

img.jpg xmin ymin xmax ymax label xmin ymin xmax ymax label .....
"""
import os
import sys
import random
import numpy as np

import torch
import torch.utils.data as data
import torchvision.transforms as transforms

from PIL import Image
#import cv2
from ..anchors import get_anchors
from fractal.loaders.bbox_transforms import resize, random_flip, random_crop, center_crop, random_distort_image
from fractal.anchors.utils import xyxy_xywh

def size_setter(seen, batch_size):
    if seen < 2000*batch_size:
        input_size = 13*32
    elif seen < 4000*batch_size:
        input_size = (random.randint(0, 3) + 13)*32
    elif seen < 8000*batch_size:
        input_size = (random.randint(0, 5)+12)*32
    elif seen < 12000*batch_size:
        input_size = (random.randint(0, 7) + 11)*32
    else:# self.seen < 20000*self.opt.batch_size:
        input_size = (random.randint(0, 9)+10)*32
    return input_size


class ListDataset(data.Dataset):
    def __init__(self, cfg, train, input_size, encode_box=True, debugger=False, multi_scale=False):
        '''
        Important Args:
          root: (str) ditectory to images.
          list_file: (str) path to index file.
          train: (boolean) train or test.
          transform: ([transforms]) image transforms.
          input_size: (int) model input size.
        '''
        self.cfg = cfg
        self.train = train
        if self.train:
            list_file = self.cfg["data"]["train_data"]
        else:
            list_file = self.cfg["data"]["valid_data"]
        self.cfg = cfg
        self.input_size = input_size
        self.multi_scale = multi_scale

        self.fnames = []
        self.boxes = []
        self.labels = []
        self.transform = transforms.Compose([transforms.ToTensor()])

        self.encode_box = encode_box
        self.debugger=debugger
        self.seen = 0

        with open(list_file) as f:
            lines = f.readlines()
            self.num_samples = len(lines)

        for line in lines:
            splited = line.strip().split()
            self.fnames.append(splited[0])
            num_boxes = (len(splited) - 1) // 5
            box = []
            label = []
            for i in range(num_boxes):
                ymin = splited[1+5*i]
                xmin = splited[2+5*i]
                ymax = splited[3+5*i]
                xmax = splited[4+5*i]
                c = splited[5+5*i]
                box.append([float(ymin),float(xmin),float(ymax),float(xmax)])
                label.append(int(c))
            self.boxes.append(torch.Tensor(box))
            self.labels.append(torch.LongTensor(label))

    def __getitem__(self, idx):
        '''Load image.
        Args:
          idx: (int) image index.
        Returns:
          img: (tensor) image tensor.
          loc_targets: (tensor) location targets.
          cls_targets: (tensor) class label targets.
        '''
        # Load image and boxes.
        if self.multi_scale:
            if self.train and idx%self.cfg["train"]["batch_size"] == 0:
                self.input_size = size_setter(self.seen, self.cfg["train"]["batch_size"])
                
        self.seen += 1
        fname = self.fnames[idx]
        #print(fname)
        #os.path.join(self.root, fname)
        img = Image.open(self.cfg["data"]["data_loc"]+fname)
        if img.mode != 'RGB':
            img = img.convert('RGB')

        boxes = self.boxes[idx].clone()

        labels = self.labels[idx]
        size = self.input_size
        
        if self.debugger:
            print("reading file:",  fname)
            print("labels:", labels)
            print("size:", size)

        # Data augmentation.
        if self.train:
            img, boxes = random_flip(img.copy(), boxes.clone())
            #img, boxes = random_crop(img, boxes)
            #print(boxes)
            img, boxes = resize(img.copy(), (self.input_size, self.input_size), boxes.clone())
            img = random_distort_image(img.copy(), self.cfg["train"]["hue"], self.cfg["train"]["saturation"], self.cfg["train"]["exposure"])
        else:
            img, boxes = resize(img, (self.input_size, self.input_size), boxes.clone())
            #img, boxes = center_crop(img, boxes, (size, size))

        img = self.transform(img)
        #print(boxes)
        #print(labels)
        if self.encode_box:
            fill_boxes = torch.zeros((500, 5))
            _, h, w = img.size()
            #print("boxes:", boxes)
            boxes = xyxy_xywh(boxes[:, [1, 0, 3, 2]], 0.5) #ctr_x, ctr_y, w, h
            #print(boxes.shape)
            #print("boxes:", boxes)
            boxes = torch.cat([labels.view(-1, 1).float(), boxes], 1)
            fill_boxes[:boxes.shape[0], :] = boxes
            return img, fill_boxes
        else:
            return img, boxes

    def __len__(self):
        return int(self.cfg["data"]["total_images"]/self.cfg["train"]["batch_size"]) * self.cfg["train"]["batch_size"]    
    @staticmethod
    def prep_image(img_loc, inp_dim):
        """
        Prepare image for inputting to the neural network. 

        Returns a Variable 
        
        img_loc: str: Location of the image
        inp_dim: tuple: resize dimension. The aspect ratio is not taken care off.
        """
        img = Image.open(img_loc)
        if img.mode != 'RGB':
            img = img.convert('RGB')
        
        shape = (img.size[1], img.size[0]) #(h, w)
        
        img_, _ = resize(img, inp_dim)
        img_tensor = transforms.ToTensor()(img_)
        img_tensor = img_tensor.unsqueeze(0)
        return img, img_tensor, shape

if __name__ == '__main__':
    import torchvision.transforms as transforms
    transform = transforms.Compose([transforms.ToTensor()])



# vim: expandtab:ts=4:sw=4
import numpy as np


def _pdist(a, b):
    """Compute pair-wise squared distance between points in `a` and `b`.

    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.

    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.

    """
    a, b = np.asarray(a), np.asarray(b)
    if len(a) == 0 or len(b) == 0:
        return np.zeros((len(a), len(b)))
    a2, b2 = np.square(a).sum(axis=1), np.square(b).sum(axis=1)
    r2 = -2. * np.dot(a, b.T) + a2[:, None] + b2[None, :]
    r2 = np.clip(r2, 0., float(np.inf))
    return r2


def _cosine_distance(a, b, data_is_normalized=False):
    """Compute pair-wise cosine distance between points in `a` and `b`.

    Parameters
    ----------
    a : array_like
        An NxM matrix of N samples of dimensionality M.
    b : array_like
        An LxM matrix of L samples of dimensionality M.
    data_is_normalized : Optional[bool]
        If True, assumes rows in a and b are unit length vectors.
        Otherwise, a and b are explicitly normalized to lenght 1.

    Returns
    -------
    ndarray
        Returns a matrix of size len(a), len(b) such that eleement (i, j)
        contains the squared distance between `a[i]` and `b[j]`.

    """
    if not data_is_normalized:
        a = np.asarray(a) / np.linalg.norm(a, axis=1, keepdims=True)
        b = np.asarray(b) / np.linalg.norm(b, axis=1, keepdims=True)
    return 1. - np.dot(a, b.T)


def _nn_euclidean_distance(x, y):
    """ Helper function for nearest neighbor distance metric (Euclidean).

    Parameters
    ----------
    x : ndarray
        A matrix of N row-vectors (sample points).
    y : ndarray
        A matrix of M row-vectors (query points).

    Returns
    -------
    ndarray
        A vector of length M that contains for each entry in `y` the
        smallest Euclidean distance to a sample in `x`.

    """
    distances = _pdist(x, y)
    return np.maximum(0.0, distances.min(axis=0))


def _nn_cosine_distance(x, y):
    """ Helper function for nearest neighbor distance metric (cosine).

    Parameters
    ----------
    x : ndarray
        A matrix of N row-vectors (sample points).
    y : ndarray
        A matrix of M row-vectors (query points).

    Returns
    -------
    ndarray
        A vector of length M that contains for each entry in `y` the
        smallest cosine distance to a sample in `x`.

    """
    distances = _cosine_distance(x, y)
    return distances.min(axis=0)


class NearestNeighborDistanceMetric(object):
    """
    A nearest neighbor distance metric that, for each target, returns
    the closest distance to any sample that has been observed so far.

    Parameters
    ----------
    metric : str
        Either "euclidean" or "cosine".
    matching_threshold: float
        The matching threshold. Samples with larger distance are considered an
        invalid match.
    budget : Optional[int]
        If not None, fix samples per class to at most this number. Removes
        the oldest samples when the budget is reached.

    Attributes
    ----------
    samples : Dict[int -> List[ndarray]]
        A dictionary that maps from target identities to the list of samples
        that have been observed so far.

    """

    def __init__(self, metric, matching_threshold, budget=None):


        if metric == "euclidean":
            self._metric = _nn_euclidean_distance
        elif metric == "cosine":
            self._metric = _nn_cosine_distance
        else:
            raise ValueError(
                "Invalid metric; must be either 'euclidean' or 'cosine'")
        self.matching_threshold = matching_threshold
        self.budget = budget
        self.samples = {}

    def partial_fit(self, features, targets, active_targets):
        """Update the distance metric with new data.

        Parameters
        ----------
        features : ndarray
            An NxM matrix of N features of dimensionality M.
        targets : ndarray
            An integer array of associated target identities.
        active_targets : List[int]
            A list of targets that are currently present in the scene.

        """
        for feature, target in zip(features, targets):
            self.samples.setdefault(target, []).append(feature)
            if self.budget is not None:
                self.samples[target] = self.samples[target][-self.budget:]
        self.samples = {k: self.samples[k] for k in active_targets}

    def distance(self, features, targets):
        """Compute distance between features and targets.

        Parameters
        ----------
        features : ndarray
            An NxM matrix of N features of dimensionality M.
        targets : List[int]
            A list of targets to match the given `features` against.

        Returns
        -------
        ndarray
            Returns a cost matrix of shape len(targets), len(features), where
            element (i, j) contains the closest squared distance between
            `targets[i]` and `features[j]`.

        """
        cost_matrix = np.zeros((len(targets), len(features)))
        for i, target in enumerate(targets):
            cost_matrix[i, :] = self._metric(self.samples[target], features)
        return cost_matrix



import torch
from PIL import Image
from .utils import inference_imshow


class error_checking():
    """ Needs a lot of brain storming.. Yet incomplete ....
    """
    def __init__(self, opt, filename, dataloader):
        self.opt = opt
        classes = [str(i) for i in range(self.opt.num_classes)]
        worst_pred = {}
        best_pred = {}
        for i in classes:
            worst_pred[i] = []
            best_pred[i] = []

        self.opt.normalize = False
        valset = dataloader(self.opt, train = "valid")
        transforms = valset.test_transformations(opt)
        incorrect_pred = open("incorrect_"+filename.strip().split("/")[3][:-4]+".txt", 'r')
        correct_pred = open("correct_"+filename.strip().split("/")[3][:-4]+".txt", 'r')
        for row in incorrect_pred:
#             print(row)
            list_ = row.strip(" \n").split(", ")
#             print(list_)
            image = Image.open(opt.data_loc + list_[0]).convert('RGB')
            image = transforms(image)
            worst_pred[list_[1]].append((image, list_[2:]))
#             print (worst_pred)
        
        self.worst_pred = worst_pred
        
        for row in correct_pred:
#             print(row)
            list_ = row.strip(" \n").split(", ")
            image = Image.open(opt.data_loc + list_[0]).convert('RGB')
            image = transforms(image)
            best_pred[list_[1]].append((image, list_[2:]))
        
        self.best_pred = best_pred
        
    def worst_prediction(self, label, num=5, imgs_per_row=4):
        '''
        Given a label aka class displays the top incorrect predictions for that class
        '''
        worst_list = self.worst_pred[label]
        if(len(worst_list)<num):
            print('Requested {} but total number of incorrect predictions is {}'.format(num, len(worst_list)))
            num = len(worst_list)
        print("Top {} worst predictions for {}:".format(num,label))
        worst_list = sorted(worst_list, key=itemgetter(1), reverse=True )
        pred_prob = [(p[0],p[1]) for c,p in worst_list[:num]]
        images = torch.stack([image.cpu() for image,_ in worst_list[:num]])
        inference_imshow(images, pred_prob, imgs_per_row = imgs_per_row)
        
    def best_prediction(self, label, num=5, imgs_per_row=4):
        '''
        Given a label aka class displays the top incorrect predictions for that class
        '''
        best_list = self.best_pred[label]
        if(len(best_list)<num):
            print('Requested {} but total number of correct predictions is {}'.format(num, len(best_list)))
            num = len(best_list)
        print("Top {} best predictions for {}:".format(num,label))
        best_list = sorted(best_list, key=itemgetter(1), reverse=True )
        pred_prob = [(p) for c,p in best_list[:num]]
        images = torch.stack([image.cpu() for image,_ in best_list[:num]])
        imshow(images, pred_prob, imgs_per_row = imgs_per_row)



import os
import torch
import torch.nn as nn
import torch.utils.model_zoo as model_zoo
from .resnext_features import resnext101_32x4d_features
from .resnext_features import resnext101_64x4d_features

__all__ = ['ResNeXt101_32x4d', 'resnext101_32x4d',
           'ResNeXt101_64x4d', 'resnext101_64x4d']

pretrained_settings = {
    'resnext101_32x4d': {
        'imagenet': {
            'url': 'http://data.lip6.fr/cadene/pretrainedmodels/resnext101_32x4d-29e315fa.pth',
            'input_space': 'RGB',
            'input_size': [3, 224, 224],
            'input_range': [0, 1],
            'mean': [0.485, 0.456, 0.406],
            'std': [0.229, 0.224, 0.225],
            'num_classes': 1000
        }
    },
    'resnext101_64x4d': {
        'imagenet': {
            'url': 'http://data.lip6.fr/cadene/pretrainedmodels/resnext101_64x4d-e77a0586.pth',
            'input_space': 'RGB',
            'input_size': [3, 224, 224],
            'input_range': [0, 1],
            'mean': [0.485, 0.456, 0.406],
            'std': [0.229, 0.224, 0.225],
            'num_classes': 1000
        }
    }
}

class ResNeXt101_32x4d(nn.Module):

    def __init__(self, num_classes=1000):
        super(ResNeXt101_32x4d, self).__init__()
        self.num_classes = num_classes
        self.features = resnext101_32x4d_features
        self.avg_pool = nn.AvgPool2d((7, 7), (1, 1))
        self.last_linear = nn.Linear(2048, num_classes)

    def logits(self, input):
        x = self.avg_pool(input)
        x = x.view(x.size(0), -1)
        x = self.last_linear(x)
        return x

    def forward(self, input):
        x = self.features(input)
        x = self.logits(x)
        return x


class ResNeXt101_64x4d(nn.Module):

    def __init__(self, num_classes=1000):
        super(ResNeXt101_64x4d, self).__init__()
        self.num_classes = num_classes
        self.features = resnext101_64x4d_features
        self.avg_pool = nn.AvgPool2d((7, 7), (1, 1))
        self.last_linear = nn.Linear(2048, num_classes)

    def logits(self, input):
        x = self.avg_pool(input)
        x = x.view(x.size(0), -1)
        x = self.last_linear(x)
        return x

    def forward(self, input):
        x = self.features(input)
        x = self.logits(x)
        return x


def resnext101_32x4d(num_classes=1000, pretrained='imagenet'):
    if pretrained == "imagenet":
        model = ResNeXt101_32x4d(num_classes=1000)
        settings = pretrained_settings['resnext101_32x4d'][pretrained]
        model.load_state_dict(model_zoo.load_url(settings['url']))

        if num_classes != 1000:
            model.last_linear = nn.Linear(2048, num_classes)
    else:
        model = ResNeXt101_32x4d(num_classes=num_classes)
    
    return model

def resnext101_64x4d(num_classes=1000, pretrained='imagenet'):
    if pretrained == "imagenet":
        model = ResNeXt101_64x4d(num_classes=1000)
        settings = pretrained_settings['resnext101_64x4d'][pretrained]
        model.load_state_dict(model_zoo.load_url(settings['url']))

        if num_classes != 1000:
            model.last_linear = nn.Linear(2048, num_classes)
    else:
        model = ResNeXt101_64x4d(num_classes=num_classes)
        
    return model


import torch
import torch.nn as nn
from torch.autograd import Variable
from functools import reduce

class LambdaBase(nn.Sequential):
    def __init__(self, fn, *args):
        super(LambdaBase, self).__init__(*args)
        self.lambda_func = fn

    def forward_prepare(self, input):
        output = []
        for module in self._modules.values():
            output.append(module(input))
        return output if output else input

class Lambda(LambdaBase):
    def forward(self, input):
        return self.lambda_func(self.forward_prepare(input))

class LambdaMap(LambdaBase):
    def forward(self, input):
        return list(map(self.lambda_func,self.forward_prepare(input)))

class LambdaReduce(LambdaBase):
    def forward(self, input):
        return reduce(self.lambda_func,self.forward_prepare(input))

resnext101_64x4d_features = nn.Sequential(#Sequential,
    nn.Conv2d(3, 64, (7, 7), (2, 2), (3, 3), 1, 1, bias = False),
    nn.BatchNorm2d(64),
    nn.ReLU(),
    nn.MaxPool2d((3, 3), (2, 2), (1, 1)),
    nn.Sequential(#Sequential,
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(64, 256, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(256),
                        nn.ReLU(),
                        nn.Conv2d(256, 256, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(256),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(256, 256, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(256),
                ),
                nn.Sequential(#Sequential,
                    nn.Conv2d(64, 256, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(256),
                ),
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(256, 256, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(256),
                        nn.ReLU(),
                        nn.Conv2d(256, 256, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(256),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(256, 256, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(256),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(256, 256, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(256),
                        nn.ReLU(),
                        nn.Conv2d(256, 256, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(256),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(256, 256, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(256),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
    ),
    nn.Sequential(#Sequential,
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(256, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                        nn.Conv2d(512, 512, (3, 3), (2, 2), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(512, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(512),
                ),
                nn.Sequential(#Sequential,
                    nn.Conv2d(256, 512, (1, 1), (2, 2), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(512),
                ),
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(512, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                        nn.Conv2d(512, 512, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(512, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(512),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(512, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                        nn.Conv2d(512, 512, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(512, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(512),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(512, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                        nn.Conv2d(512, 512, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(512),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(512, 512, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(512),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
    ),
    nn.Sequential(#Sequential,
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(512, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (2, 2), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                nn.Sequential(#Sequential,
                    nn.Conv2d(512, 1024, (1, 1), (2, 2), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                        nn.Conv2d(1024, 1024, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(1024),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(1024, 1024, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(1024),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
    ),
    nn.Sequential(#Sequential,
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(1024, 2048, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(2048),
                        nn.ReLU(),
                        nn.Conv2d(2048, 2048, (3, 3), (2, 2), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(2048),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(2048, 2048, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(2048),
                ),
                nn.Sequential(#Sequential,
                    nn.Conv2d(1024, 2048, (1, 1), (2, 2), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(2048),
                ),
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(2048, 2048, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(2048),
                        nn.ReLU(),
                        nn.Conv2d(2048, 2048, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(2048),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(2048, 2048, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(2048),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
        nn.Sequential(#Sequential,
            LambdaMap(lambda x: x, #ConcatTable,
                nn.Sequential(#Sequential,
                    nn.Sequential(#Sequential,
                        nn.Conv2d(2048, 2048, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                        nn.BatchNorm2d(2048),
                        nn.ReLU(),
                        nn.Conv2d(2048, 2048, (3, 3), (1, 1), (1, 1), 1, 64, bias = False),
                        nn.BatchNorm2d(2048),
                        nn.ReLU(),
                    ),
                    nn.Conv2d(2048, 2048, (1, 1), (1, 1), (0, 0), 1, 1, bias = False),
                    nn.BatchNorm2d(2048),
                ),
                Lambda(lambda x: x), #Identity,
            ),
            LambdaReduce(lambda x, y: x + y), #CAddTable,
            nn.ReLU(),
        ),
    )
)


import torch
import torch.nn as nn
import torch.nn.functional as F

def get_loss_function(opt):
    if opt.loss_type == "cross_entropy_loss":
        return torch.nn.CrossEntropyLoss(weight=opt.class_weight)
    elif opt.loss_type == "focal_loss":
        return FocalLoss(opt.num_classes, opt.use_gpu)
    elif opt.loss_type == "multilabel_BCELoss":
        return torch.nn.BCEWithLogitsLoss(reduction="sum")
    else:
        raise NotImplementedError


        
class FocalLoss(nn.Module):
    def __init__(self,
                 classes,
                 use_gpu,
                 focusing_param=2.0,
                 balance_param=0.25):
        super().__init__()
        self.use_gpu = use_gpu
        self.focusing_param = focusing_param
        self.balance_param = balance_param
        self.classes = classes
        
    def forward(self, x, y):
        batch_size, next_best = y.size()[0], y.size()[1]
        sigmoid_p = torch.sigmoid(x)
        zeros = torch.zeros(sigmoid_p.shape)
        if self.use_gpu:
            zeros = zeros.cuda()
        y = y.float()
        
        pos_p_sub = ((y >= sigmoid_p).float() * (y-sigmoid_p)) + ((y < sigmoid_p).float() * zeros)
        neg_p_sub = ((y >= zeros).float() * zeros) + ((y <= zeros).float() * sigmoid_p)
        
        ce = (-1) * self.balance_param * (pos_p_sub ** self.focusing_param) * torch.log(torch.clamp(sigmoid_p, 1e-4, 1.0)) -(1-self.balance_param) * (neg_p_sub ** self.focusing_param) * torch.log(torch.clamp(1.0-sigmoid_p, 1e-4, 1.0))
        #pos_samples = float(batch_size * next_best)
        return ce.sum()/y.sum()
        